
catch {
   package require img::gif
   package require img::jpeg
   package require img::png
   package require compass_utils::c
}

package require textutil
package require math::constants
package require compass_utils

namespace eval PDFWriter::PDF {
    variable pdffile
    variable pdfhandle ""
    variable pagehandle
    variable fonts
    variable xmax
    variable ymax
    variable cacheimages
    variable tcl_encoding
    variable pdf_encoding
    
    namespace export *
}

proc PDFWriter::PDF::cm2pt { cm } {
    return [expr $cm*72/2.54]
}

proc PDFWriter::PDF::pt2cm { pt } {
    return [expr $pt*2.54/72.0]
}

proc PDFWriter::PDF::GiveEncoding {} {
    variable tcl_encoding

    return $tcl_encoding
}

proc PDFWriter::PDF::Init { { papertype A4 } { file "" } } {

    return [InitLibrary -papertype $papertype -file $file]
}

proc PDFWriter::PDF::InitLibrary { args } {
    variable pdffile
    variable pdfhandle
    variable pagehandle
    variable fonts
    variable xmax
    variable ymax
    variable cacheimages
    variable cachefonts
    variable local_link_stack
    variable tcl_encoding
    variable pdf_encoding
    variable fontsdir
    variable use_utf8

    set optional {
	{ -papertype type A4 }
	{ -file file "" }
	{ -fontsdir directory "" }
	{ -filename filename "" }
	{ -use_utf8 boolean 1 }
     }
     set compulsory ""
    parse_args $optional $compulsory $args
    
    if { !$use_utf8 } {
	set tcl_encoding cp1252
	set pdf_encoding WinAnsiEncoding
    } else {
	set tcl_encoding utf-8
	set pdf_encoding UTF-8
	#set pdf_encoding Modern-UTF8-H
	#set pdf_encoding Ancient-UTF8-H
    }
    set pdffile $file
    
    array unset cacheimages
    set pdfhandle [hpdf::new]
    $pdfhandle setCompressionMode all
    
    #PDF_set_parameter $pdfhandle serial "GiD"

    if { $pdffile eq "" } {
	set ic 0
	while { $ic < 50 } {
	    set tmpdir /windows/temp
	    catch { set tmpdir [GiveTempDir] }
	    if { $filename eq "" } {
		set pdffile [file join $tmpdir tmpfile$ic.pdf]
	    } elseif { [file pathtype $filename] eq "absolute" } {
		set pdffile $filename
	    } else {
		if { ![regexp {^UNIQUE_KEY_(.*)} $filename {} filename] } {
		    if { [lindex [split $filename _] end] eq [expr {$ic - 1}] } {
		        set filename [join [lrange [split $filename _] 0 end-1] _]
		    } 
		    set filename [join [list $filename $ic] _]
		}
		set pdffile [file join $tmpdir $filename.pdf]
	    }
	    if { [file exists $pdffile] && [clock seconds] - [file mtime $pdffile] < 30 } {
		incr ic; continue
	    }
	    set err [catch { $pdfhandle saveToFile $pdffile } err_txt]
	    if { !$err } { break } else { W $err_txt}
	    incr ic
	}
	if { $ic == 50 } {
	    error [_ "Error: could not open temporal file"]
	}
    } else {
	set err [catch { $pdfhandle saveToFile $pdffile }]
	if { $err } {
	    error [_ "Error: could not open file '%s'" $pdffile]
	}
    }

    switch -glob -- $papertype {
	A5 {
	    set xmax [cm2pt 14.85]
	    set ymax [cm2pt 21]
	    set bigmarginh [cm2pt 1.5]
	    set bigmarginv [cm2pt 1.5]
	    set ysep 5
	}
	A5L {
	    set xmaxL [cm2pt 21]
	    set ymaxL [cm2pt 14.85]
	    set bigmarginh [cm2pt 1.5]
	    set bigmarginv [cm2pt 1.5]
	    set ysep 5
	}
	A4 {
	    set xmax [cm2pt 21]
	    set ymax [cm2pt 29.7]
	    set bigmarginh [cm2pt 1.5]
	    set bigmarginv [cm2pt 1.5]
	    set ysep 5
	}
	A4L {
	    set xmax [cm2pt 29.7]
	    set ymax [cm2pt 21]
	    set bigmarginh [cm2pt 2.5]
	    set bigmarginv [cm2pt 1.5]
	    set ysep 5
	}
	A3 {
	    set xmax [cm2pt 29.7]
	    set ymax [cm2pt 42.0]
	    set bigmarginh [cm2pt 1.5]
	    set bigmarginv [cm2pt 1.5]
	    set ysep 5
	}
	A3L {
	    set xmax [cm2pt 42.0]
	    set ymax [cm2pt 29.7]
	    set bigmarginh [cm2pt 1.5]
	    set bigmarginv [cm2pt 1.5]
	    set ysep 5
	}
	userdefined* {
	    set xmax [cm2pt [lindex $papertype 1]]
	    set ymax [cm2pt [lindex $papertype 2]]
	    if { [llength $papertype] > 3 } {
		set bigmarginh [cm2pt [lindex $papertype 3]]
		set bigmarginv [cm2pt [lindex $papertype 4]]
	    } else {
		set bigmarginh [cm2pt 1.5]
		set bigmarginv [cm2pt 1.5]
	    }
	    set ysep 5
	}
	default {
	    error "error unknown paper type"
	}
    }

    $pdfhandle setInfoAttr "creator" "GiD"
    $pdfhandle setInfoAttr "producer" "pdfwriter"
    $pdfhandle setInfoAttr "author" "GiD Team"
    $pdfhandle setInfoAttr "title" "GiD"

    #PDF_set_border_style $pdfhandle solid 0

    set pagehandle [$pdfhandle addPage]
    $pagehandle setWidth $xmax
    $pagehandle setHeight $ymax
    
    set  local_link_stack ""
    array unset fonts
    set cachefonts ""

    if { $use_utf8 } {
	$pdfhandle UseUTFEncodings
	$pdfhandle setCurrentEncoder $pdf_encoding
	# osVersion == 11.0.0 ... 11.4.2 == Mac OS X 10.7.5
	# osVersion == 12.0.0 ... 12.5.0 == Mac OS X 10.8
	# osVersion == 13.0.0 ... 13.4.0 == Mac OS X 10.9.5
	set os_major_version [ lindex [ split $::tcl_platform(osVersion) .] 0]
	# we could have used X11 fonts also for mac os x < 10.8
	# but better to use the same look on all Mac Os X platforms
	# && ( $os_major_version >= 12)
	if { ( $::tcl_platform(platform) eq "windows") || ( $::tcl_platform(os) == "Darwin")} {
	    if { $::tcl_platform(platform) eq "windows" } {
		set nl [list DejaVuSans-ExtraLight tahoma verdana arial]
		set no [list verdana tahoma arial]
		set b [list verdanab tahomabd arialbd]
		set i [list verdanai tahomabd ariali]
		set bi [list verdanaz tahomabd arialbi]
		set fontsdir [ file join $::env(windir) fonts]
	    } else {
		set nl [list DejaVuSans-ExtraLight Tahoma Verdana Arial]
		set no [list Verdana Tahoma Arial]
		set b [list "Verdana Bold" "Tahoma Bold" "Arial Bold"]
		set i [list "Verdana Italic" "Tahoma Bold" "Arial Italic"]
		set bi [list "Verdana Bold Italic" "Tahoma Bold" "Arial Bold Italic"]
		set fontsdir /Library/Fonts
	    }
	    foreach "n fileList" [list "" $no "-Bold" $b "-Oblique" $i "-BoldOblique" $bi "-Light" $nl] {
		if { ![info exists fonts(Helvetica$n)] } {
		    foreach fl $fileList {
		        set file [file join $fontsdir $fl.ttf]
		        if { ![file exists $file] } { continue }
		        if { ![dict exists $cachefonts files $file] } {
		            set f [$pdfhandle loadTTFontFromFile $file 1]
		            dict set cachefonts files $file $f
		        } else {
		            set f [dict get $cachefonts files $file]
		        }
		        break
		    }
		    set fonts(Helvetica$n) [$pdfhandle getFont $f $pdf_encoding]
		}
	    }
	    if { $::tcl_platform(platform) eq "windows" } {
		set no [list cour verdana tahoma arial]
		set b [list courbd verdanab tahomabd arialbd]
		set i [list couri verdanai tahomabd ariali]
		set bi [list courbi verdanaz tahomabd arialbi]
		set fontsdir [ file join $::env(windir) fonts]
	    } else {
		set no [list "Courier New" "Andale Mono" Verdana Tahoma Arial]
		set b [list "Courier New Bold" "Andale Mono" "Verdana Bold" "Tahoma Bold" "Arial Bold"]
		set i [list "Courier New Italic" "Andale Mono" "Verdana Italic" "Tahoma Bold" "Arial Italic"]
		set bi [list "Courier New Bold Italic" "Andale Mono" "Verdana Bold Italic" "Tahoma Bold" "Arial Bold Italic"]
		set fontsdir /Library/Fonts
	    }
	    foreach "n fileList" [list "" $no "-Bold" $b "-Oblique" $i "-BoldOblique" $bi] {
		if { ![info exists fonts(Courier$n)] } {
		    foreach fl $fileList {
		        set file [file join $fontsdir $fl.ttf]
		        if { ![file exists $file] } { continue }
		        if { ![dict exists $cachefonts files $file] } {
		            set f [$pdfhandle loadTTFontFromFile $file 1]
		            dict set cachefonts files $file $f
		        } else {
		            set f [dict get $cachefonts files $file]
		        }
		        break
		    }
		    set fonts(Courier$n) [$pdfhandle getFont $f $pdf_encoding]
		}
	    }
	} else {
	    set families [list -family Ubuntu -family FreeSans -family Arial]
	    foreach "n weight slant" [list "" normal roman "-Bold" bold roman "-Oblique" normal italic \
		    "-BoldOblique" bold italic "-Light" light roman] {
		if { ![info exists fonts(Helvetica$n)] } {
		    set file [cu::find_font_file {*}$families -weight $weight -slant $slant]
		    if { ![dict exists $cachefonts files $file] } {
		        set f [$pdfhandle loadTTFontFromFile $file 1]
		        dict set cachefonts files $file $f
		    } else {
		        set f [dict get $cachefonts files $file]
		    }
		    set fonts(Helvetica$n) [$pdfhandle getFont $f $pdf_encoding]
		}
		if { ![info exists fonts(Courier$n)] } {
		    set file [cu::find_font_file {*}$families -weight $weight -slant $slant -fixed 1]
		    if { ![dict exists $cachefonts files $file] } {
		        set f [$pdfhandle loadTTFontFromFile $ile 1]
		        dict set cachefonts files $file $f
		    }
		    set fonts(Courier$n) [$pdfhandle getFont $f $pdf_encoding]
		}
	    }
	}
    } else {
#     set fonts(Helvetica) [PDF_findfont $pdfhandle "Tahoma" "host" 0]
#     set fonts(Helvetica-Bold) [PDF_findfont $pdfhandle "Tahoma Bold" "host" 0]
	set fonts(Helvetica) [$pdfhandle getFont "Helvetica" "WinAnsiEncoding"]
	set fonts(Helvetica-Bold) [$pdfhandle getFont "Helvetica-Bold" "WinAnsiEncoding"]
	set fonts(Helvetica-Oblique) [$pdfhandle getFont "Helvetica-Oblique" "WinAnsiEncoding"]
	set fonts(Helvetica-BoldOblique) [$pdfhandle getFont "Helvetica-BoldOblique" "WinAnsiEncoding"]
    }

    if { $::tcl_platform(platform) eq "windows" } {
	if 0 {
	    set err [catch {
		    set fontname {\Windows\Fonts\tahoma.ttf}
		    set font [$pdfhandle loadTTFontFromFile $fontname false]
		    
		    set fontname {\Windows\Fonts\arial.ttf}
		    set font_Arial [$pdfhandle loadTTFontFromFile $fontname false]
		}]
	    if { $err } {
		set font Helvetica
		set font_Arial Times-Roman
	    }
	} else {
	    set font Helvetica
	    set font_Arial Times-Roman
	}
    } else {
	set font Helvetica
	set font_Arial Times-Roman
    }
    if { !$use_utf8 } {
	set fonts(Times-Roman) [$pdfhandle getFont $font "WinAnsiEncoding"]
	set fonts(Times-Bold) [$pdfhandle getFont $font "WinAnsiEncoding"]
	set fonts(Times-Italic) [$pdfhandle getFont $font "WinAnsiEncoding"]
	set fonts(Times-BoldItalic) [$pdfhandle getFont $font "WinAnsiEncoding"]
#     set fonts(Helvetica-Oblique) [PDF_findfont $pdfhandle "Tahoma" "WinAnsiEncoding" 0]
#     set fonts(Helvetica-BoldOblique) [PDF_findfont $pdfhandle "Tahoma" "WinAnsiEncoding" 0]
    
	set fonts(Courier-Bold) [$pdfhandle getFont "Courier-Bold" "WinAnsiEncoding"]
	set fonts(Courier) [$pdfhandle getFont "Courier" "WinAnsiEncoding"]
    } else {
	
	set d [dict create Roman "" Bold "-Bold" Italic "-Oblique" BoldItalic "-BoldOblique"]
	dict for "n v" $d {
	    set fonts(Times-$n) $fonts(Helvetica$v)
	}
    }
    set fonts(Symbol) [$pdfhandle getFont "Symbol" ""]

    set fonts(Arial) [$pdfhandle getFont $font_Arial "WinAnsiEncoding"]
    #set fonts(Arial-Bold) [PDF_findfont $pdfhandle "Arial,Bold" "WinAnsiEncoding" 

    if 0 {
	# this code crashes in some computers
	set fonts(Times-Roman) [PDF_findfont $pdfhandle "Times-Roman" "WinAnsiEncoding" 0]
	set fonts(Times-Bold) [PDF_findfont $pdfhandle "Times-Bold" "WinAnsiEncoding" 0]
	set fonts(Times-Italic) [PDF_findfont $pdfhandle "Times-Italic" "WinAnsiEncoding" 0]
	set fonts(Times-BoldItalic) [PDF_findfont $pdfhandle "Times-BoldItalic" "WinAnsiEncoding" 0]
	set fonts(Helvetica-Oblique) [PDF_findfont $pdfhandle "Helvetica-Oblique" "WinAnsiEncoding" 0]
	set fonts(Helvetica-BoldOblique) [PDF_findfont $pdfhandle "Helvetica-BoldOblique" "WinAnsiEncoding" 0]
	set fonts(Courier-Bold) [PDF_findfont $pdfhandle "Courier-Bold" "WinAnsiEncoding" 0]
	set fonts(Courier) [PDF_findfont $pdfhandle "Courier" "WinAnsiEncoding" 0]
	set fonts(Symbol) [PDF_findfont $pdfhandle "Symbol" "WinAnsiEncoding" 0]
    }

    #PDF_set_parameter $pdfhandle nativeunicode true

    return [list $xmax $ymax $bigmarginh $bigmarginv $ysep]
}

proc PDFWriter::PDF::SetFontAndSize { font size } {
    variable pdfhandle
    variable pagehandle
    variable pdf_encoding
    variable fonts
    variable fontsdir
    variable use_utf8
    variable cachefonts
    
    if { ![info exists fonts($font)] } {
	if { $fontsdir eq "" } {
	    error "error: fontsdir is not defined"
	}
	set file [file join $fontsdir $font.ttf]
	if { ![file exists $file] && $::tcl_platform(platform) ne "windows" } {
	    set file [cu::find_font_file -family $font]
	}
	if { ![dict exists $cachefonts files $file] } {
	    set err [catch { $pdfhandle loadTTFontFromFile [file join $fontsdir $font.ttf] true } f]
	    if { $err } {
		error [_ "Error loading font '%s'. (%s)" $font $f]
	    }
	    dict set cachefonts files $file $f
	}
	set fonts($font) [$pdfhandle getFont [dict get $cachefonts files $file] $pdf_encoding]
    }
    $pagehandle setFontAndSize $fonts($font) $size
}

proc PDFWriter::PDF::GiveTempDir {} {
    variable tmpdir

    if { [info exists tmpdir] } {
	return $tmpdir
    }

    if { [info exists ::env(TEMP)] && [file isdir $::env(TEMP)] } {
	set tmpdir $::env(TEMP)
    } elseif { [info exists ::env(TMP)] && [file isdir $::env(TMP)] } {
	set tmpdir $::env(TMP)
    } elseif { [info exists ::env(WINDIR)] && [file isdir [file join $::env(WINDIR) temp]] } {
	set tmpdir [file join $::env(WINDIR) temp]
    } elseif { ( $::tcl_platform(os) == "Darwin") && [file isdir /var/tmp] } { 
	set tmpdir /var/tmp
    } elseif { [file isdir /tmp] } {
	set tmpdir /tmp
    } else {
	set tmpdir /
    }
    return $tmpdir
}

proc PDFWriter::PDF::AddBookmark { text parent } {
    variable pdfhandle
    variable pagehandle
    variable pdf_encoding
    
#     if { $parent } {
#         set open 1
#     } else { set open 0 }

    set open 0

    if { $parent == 0 } { set parent "" }
    set out_handle [$pdfhandle createOutline $parent [_encode_text $text ""] \
	    [$pdfhandle getEncoder $pdf_encoding]]
    $out_handle setOpened 0
    $out_handle setDestination [$pagehandle createDestination]
    return $out_handle
}

proc PDFWriter::PDF::NewPage { { papertype "" } } {
    variable pdfhandle
    variable pagehandle
    variable xmax
    variable ymax

    switch -glob -- $papertype {
	"" { 
	    lassign [list $xmax $ymax] xmaxL ymaxL
	}
	A5 {
	    set xmaxL [cm2pt 14.85]
	    set ymaxL [cm2pt 21]
	}
	A5L {
	    set xmaxL [cm2pt 21]
	    set ymaxL [cm2pt 14.85]
	}
	A4 {
	    set xmaxL [cm2pt 21]
	    set ymaxL [cm2pt 29.7]
	}
	A4L {
	    set xmaxL [cm2pt 29.7]
	    set ymaxL [cm2pt 21]
	}
	A3 {
	    set xmaxL [cm2pt 29.7]
	    set ymaxL [cm2pt 42.0]
	}
	A3L {
	    set xmaxL [cm2pt 42.0]
	    set ymaxL [cm2pt 29.7]
	}
	userdefined* {
	    set xmaxL [cm2pt [lindex $papertype 1]]
	    set ymaxL [cm2pt [lindex $papertype 2]]
	}
	default {
	    error "error unknown paper type"
	}
    }
    set pagehandle [$pdfhandle addPage]
    $pagehandle setWidth $xmaxL
    $pagehandle setHeight $ymaxL
    
    return [list $xmaxL $ymaxL]
}


# TCL font is supposed to be in the format: -size .. -family ..
proc PDFWriter::PDF::TclfontToPDF { font sizedefault } {
    
    set fail_tk [catch { package present Tk }]

    set ipos [lsearch $font -size]
    if { $ipos != -1 } {
	incr ipos
	if { !$fail_tk } { 
	    set size [font actual $font -size]
	    if { $size < 0 } {
		set size [expr {round(abs($size)/[tk scaling])}]
	    }
	} else {
	    dict set font -size [expr {1*[dict get $font -size]}]
	    #temporary solution to simulate tk scaling in remote sessions
	    set size [expr {round(abs([dict get $font -size]/1.33))}]
	}
	set sizefac [expr double($size)/double($sizedefault)]
    } else {
	set sizefac 1.0
    }
    set ipos [lsearch $font -family]
    if { $ipos != -1 } {
	incr ipos
	set family [lindex $font $ipos]
	if { [string tolower $family] == "symbol" } {
	    return [list Symbol $sizefac]
	}
    }
    if { [lsearch $font bold] != -1 } {
	if { [lsearch $font italic] != -1 } {
	    return [list Helvetica-BoldOblique $sizefac]
	} else {
	    return [list Helvetica-Bold $sizefac]
	}
    } elseif { [lsearch $font italic] != -1 } {
	return [list Helvetica-Oblique $sizefac]
    } else {
	return [list Helvetica $sizefac]
    }


#     set size [font actual $font -size]
#     set sizefac [expr double($size)/double($sizedefault)]
# 
#     set family [font actual $font -family]
#     if { [string tolower $family] == "symbol" } {
#         return [list Symbol $sizefac]
#     }
#     set weight [font actual $font -weight]
#     set slant [font actual $font -slant]
#     if { $weight eq "bold" } {
#         if { $slant eq "italic" } {
#             return [list Helvetica-BoldOblique $sizefac]
#         } else {
#             return [list Helvetica-Bold $sizefac]   
#         }
#     } elseif { $slant eq "italic" } {
#         return [list Helvetica-Oblique $sizefac]
#     } else {
#         return [list Helvetica $sizefac]
#     }
}

proc PDFWriter::PDF::is_png { data } {
    return [string equal [string range $data 0 7] "\x89\x50\x4e\x47\xd\xa\x1a\xa"]
}

proc PDFWriter::PDF::InsertImageNoDisk { img x y anchor { scale 1 } } {
    return [InsertImageFromData [cu::base64 decode [$img data -format png]] $x $y $anchor $scale]
}

proc PDFWriter::PDF::InsertImageFromData { imgdata x y anchor { scale 1 } } {
    variable pdfhandle
    
    if { [string range $imgdata 0 2] == "GIF" } {
	set img [image create photo -data $imgdata]
	set imgdata [cu::base64 decode [$img data -format png]]
	image delete $img
	set type png
    } elseif { [string range $imgdata 1 3] == "PNG" } {
	set type png
    } else {
	set type jpeg
    }
    switch $type {
	png {
	    set image [$pdfhandle loadPngImageFromMem $imgdata]
	}
	jpeg {
	    set image [$pdfhandle loadJpegImageFromMem $imgdata]
	}
    }
    return [InsertImageFromPDFImage $image $x $y $anchor $scale]
}


proc PDFWriter::PDF::_parse_scale { scale width height } {
    
    if { [regexp {(?i)(h|v)=([-0-9.]+)cm} $scale {} orient cms] } {
	switch $orient {
	    h { set scale [expr [cm2pt $cms]/double($width)] }
	    v { set scale [expr [cm2pt $cms]/double($height)] }
	}
    }
    if { [regexp {(?i)(h|v)=([-0-9.]+)pt} $scale {} orient pts] } {
	switch $orient {
	    h { set scale [expr $pts/double($width)] }
	    v { set scale [expr $pts/double($height)] }
	}
    }
    return $scale
}

# scale can be either a number or: h=2.3cm or v=1cm or h=234pt or two of
# the previous separated by comma

proc PDFWriter::PDF::InsertImageFromPDFImage { PDFImage x y anchor { scale 1 } } {
    variable pdfhandle
    variable pagehandle
	
    set width [$PDFImage getWidth]
    set height [$PDFImage getHeight]

    if { [llength [split $scale ","]] == 2 } {
	lassign [split $scale ","] scalex scaley
	foreach i [list scalex scaley] {
	    set $i [_parse_scale [set $i] $width $height]
	}
    } else {
	set scale [_parse_scale $scale $width $height]
	lassign [list $scale $scale] scalex scaley
    }
    set width [expr {$width*$scalex}]
    set height [expr {$height*$scaley}]
    
    if { $anchor eq "center" } { set anchor "" }
    if { [string first n $anchor] != -1 } {
	set y [expr {$y-$height}]
    } elseif { [string first s $anchor] != -1 } {
	# nothing
    } else {
	set y [expr {$y-$height/2.0}]
    }
    if { [string first w $anchor] != -1 } {
	# nothing
    } elseif { [string first e $anchor] != -1 } {
	set x [expr {$x-$width}]
    } else {
	set x [expr {$x-$width/2.0}]
    }
    
#     if { $scalex != $scaley } {
#         set fac [expr {$scaley/double($scalex)}]
#         $pagehandle gSave
#         $pagehandle concat 1.0 0 0 $fac 0 0
#         set y [expr {$y/$fac}]
#     }
    $pagehandle drawImage $PDFImage $x $y $width $height
    
#     if { $scalex != $scaley } {
#         $pagehandle gRestore
#     }
    return [list $width $height]
}

proc PDFWriter::PDF::InsertImage { imagename x y anchor { scale 1 } } {
    
    return [InsertImageFromData [cu::base64 decode [set Images::${imagename}_data]] $x $y $anchor $scale]
}


proc PDFWriter::PDF::InsertImg { img x y anchor { scale 1 } } {
    return [InsertImageFromData [cu::base64 decode [$img data -format png]] $x $y $anchor $scale]
}

proc PDFWriter::PDF::_trycopyingfile { file } {

    # useful for tclkit
    if { ![file exists $file] } { return "" }

    if { [info exists ::env(APPDATA)] } {
	set dir $::env(APPDATA)
    } elseif { [info exists ::env(TEMP)] } {
	set dir $::env(TEMP)
    } elseif { [file isdirectory /WINDOWS/Temp] } {
	set dir /WINDOWS/Temp
    } else { set dir c:/ }

    set base [file root [file tail [info nameofexecutable]]]
    if { $base == "" } {
	set base [file root [file tail $::argv0]]
    }
    if { $base != "" } {
	set dir [file join $dir $base]
	file mkdir $dir
    }
    if { ![file exists [file join $dir [file tail $file]]] } {
	file copy $file $dir
    }
    return [file join $dir [file tail $file]]
}

proc PDFWriter::PDF::InsertImageFile { file x y anchor { scale 1 } } {
    variable pdfhandle
    variable cacheimages

    switch [file ext $file] {
	.gif { set type gif }
	.jpg - .jpeg { set type jpeg }
	.png { set type png }
	default {
	    error "unknown extension for image file '$file'"
	    return
	}
    }
    if { [info exists cacheimages($file)] } {
	set PDFImage $cacheimages($file)
    } else {
	switch $type {
	    gif {
		set img [image create photo -file $file]
		set imgdata [$img data -format png]
		catch { set imgdata [cu::base64 decode $imgdata] }
		image delete $img
		set PDFImage [$pdfhandle loadPngImageFromMem $imgdata]
	    }
	    png {
		set img [image create photo -file $file]
		set imgdata [$img data -format png]
		catch { set imgdata [cu::base64 decode $imgdata] }
		image delete $img
		set PDFImage [$pdfhandle loadPngImageFromMem $imgdata]
	    }
	    jpeg {
		set img [image create photo -file $file]
		set imgdata [$img data -format jpeg]
		catch { set imgdata [cu::base64 decode $imgdata] }
		image delete $img
		set PDFImage [$pdfhandle loadJpegImageFromMem $imgdata]
	    }
	}
	set cacheimages($file) $PDFImage
    }
    return [InsertImageFromPDFImage $PDFImage $x $y $anchor $scale]
}

proc PDFWriter::PDF::ImageFileSize { file } {
    variable pdfhandle
    
    switch [file ext $file] {
	.gif { set type gif }
	.jpg - .jpeg { set type jpeg }
	.png { set type png }
	default {
	    error "unknown extension for image file '$file'"
	    return
	}
    }
    if { [info exists cacheimages($file)] } {
	set PDFImage $cacheimages($file)
    } else {
	switch $type {
	    gif {
		set img [image create photo -file $file]
		set imgdata [cu::base64 decode [$img data -format png]]
		image delete $img
		set PDFImage [$pdfhandle loadPngImageFromMem $imgdata]
	    }
	    png {
		set img [image create photo -file $file]
		set imgdata [cu::base64 decode [$img data -format png]]
		image delete $img
		set PDFImage [$pdfhandle loadPngImageFromMem $imgdata]
	    }
	    jpeg {
		set img [image create photo -file $file]
		set imgdata [cu::base64 decode [$img data -format jpeg]]
		image delete $img
		set PDFImage [$pdfhandle loadJpegImageFromMem $imgdata]
	    }
	}
	set cacheimages($file) $PDFImage
    }
    set width [$PDFImage getWidth]
    set height [$PDFImage getHeight]

    return [list $width $height]
}

proc PDFWriter::PDF::_SetColorDashLineWidth { color dash lwidth } {
    global __tcl3dColorList
    variable pdfhandle
    variable pagehandle
	
    if { $color ne "" } {
	#PREVIOUS CODE:
	#lassign [winfo rgb . $color] c1 c2 c3
	#set c1 [expr $c1/65535.0]
	#set c2 [expr $c2/65535.0]
	#set c3 [expr $c3/65535.0]
	
	#REPLACED BY THE FOLLOWING:
	if {[string index $color 0] eq "#" } {
	    scan $color "#%2x%2x%2x" r g b
	    set c1 [expr $r/255.0]
	    set c2 [expr $g/255.0]
	    set c3 [expr $b/255.0]
	} else {
	    set ind [lsearch -nocase $PDFWriter::__tcl3dColorList $color]
	    if { $ind >= 0 } {
		lassign [lrange $PDFWriter::__tcl3dColorList [expr $ind+1] [expr $ind+3]] r g b
		set c1 [expr $r/255.0]
		set c2 [expr $g/255.0]
		set c3 [expr $b/255.0]
	    } else {
		error "Could not find color $color"
	    }
	}
	$pagehandle setRGBFill $c1 $c2 $c3
	$pagehandle setRGBStroke $c1 $c2 $c3
    }
    if { ![string equal $dash 0] && $dash ne "" } { 
	if { [llength $dash] == 2 } {
	    $pagehandle setDash $dash 0
	} else {
	    $pagehandle setDash [list 5 5] 0
	}
    }
    if { $lwidth ne "" && $lwidth != 0 } {
	$pagehandle setLineWidth $lwidth
    }
}

proc PDFWriter::PDF::_ResetColorDashLineWidth {} {
    variable pdfhandle
    variable pagehandle
    
    $pagehandle setRGBFill 0 0 0
    $pagehandle setRGBStroke 0 0 0
    $pagehandle setLineWidth 1
    $pagehandle setDash "" 0
}

proc PDFWriter::PDF::CreateLine { x1 y1 x2 y2 lwidth color { dash 0 } } {
    variable pdfhandle
    variable pagehandle
    
    _SetColorDashLineWidth $color $dash $lwidth
    $pagehandle moveTo $x1 $y1
    $pagehandle lineTo $x2 $y2
    $pagehandle stroke
    _ResetColorDashLineWidth
}

proc PDFWriter::PDF::CreateLineC { coords lwidth color { dash 0 } } {
    variable pdfhandle
    variable pagehandle
    
    _SetColorDashLineWidth $color $dash $lwidth
    $pagehandle moveTo {*}[lrange $coords 0 1]
    foreach "x y" [lrange $coords 2 end] {
	$pagehandle lineTo $x $y
    }
    $pagehandle stroke
    _ResetColorDashLineWidth
}

proc PDFWriter::PDF::CreateLineArrow { x1 y1 x2 y2 arrowtype lwidth color { dash 0 } } {
    variable pdfhandle
    variable pagehandle
    
    set PI [expr {atan2(0,-1)}]
    _SetColorDashLineWidth $color $dash $lwidth
    
    $pagehandle moveTo $x1 $y1
    $pagehandle lineTo $x2 $y2
    $pagehandle stroke

    if { $arrowtype in "first both" } {
	$pagehandle gSave
	set angle [expr {atan2($y1-$y2,$x1-$x2)}]
	$pagehandle concat [expr {cos($angle)}] [expr {sin($angle)}] \
	    [expr {-sin($angle)}] [expr {cos($angle)}] $x1 $y1
	$pagehandle moveTo 0 0
	$pagehandle lineTo -6 -3
	$pagehandle lineTo -6 +3
	$pagehandle closePath
	$pagehandle fill
	$pagehandle gRestore
    }
    if { $arrowtype == "last" ||$arrowtype == "both" } {
	$pagehandle gSave
	set angle [expr {atan2($y2-$y1,$x2-$x1)}]
	$pagehandle concat [expr {cos($angle)}] [expr {sin($angle)}] \
	    [expr {-sin($angle)}]  [expr {cos($angle)}] $x2 $y2
	$pagehandle moveTo 0 0
	$pagehandle lineTo -6 -3
	$pagehandle lineTo -6 +3
	$pagehandle closePath
	$pagehandle fill
	$pagehandle gRestore
    }
    _ResetColorDashLineWidth $pagehandle
}

proc PDFWriter::PDF::CreateBezierD2 { coordList lwidth color { dash 0 } { isfilled 0 } } {
    
#     set idx 1
#     for { set i 0 } { $i < [llength $coordList] } { incr i 2 } {
#         lassign [lrange $coordList $i $i+1] x y
#         CreateCircle $x $y 2 1 blue 0 1
#         WriteText $idx [expr {$x+2}] [expr {$y+2}] Arial 12 sw blue
#         incr idx
#     }

    set points3 [lrange $coordList 0 1]
    for { set i 0 } { $i < [llength $coordList]-2 } { incr i 4 } {
	for { set k 1 } { $k < 3 } { incr k } {
	    set Pkm1 [lrange $coordList [expr {$i+2*$k-2}] [expr {$i+2*$k-1}]]
	    set Pk [lrange $coordList [expr {$i+2*$k}] [expr {$i+2*$k+1}]]
	    lappend points3 {*}[m::add [m::scale [expr {$k/3.0}] $Pkm1] \
		    [m::scale [expr {1-$k/3.0}] $Pk]]
	}
	lappend points3 {*}[lrange $coordList $i+4 $i+5]
    }
    CreateBezierD3 $points3 $lwidth $color $dash $isfilled
}

proc PDFWriter::PDF::CreateBezierD3 { coordList lwidth color { dash 0 } { isfilled 0 } } {
    variable pdfhandle
    variable pagehandle
    
    _SetColorDashLineWidth $color $dash $lwidth
    $pagehandle moveTo {*}[lrange $coordList 0 1]
    for { set i 2 } { $i < [llength $coordList] } { incr i 6 } {
	if { $i+6 > [llength $coordList] } {
	    break
	}
	$pagehandle curveTo {*}[lrange $coordList $i $i+5]
    }
    if { !$isfilled } {
	$pagehandle stroke
    } else {
	$pagehandle fill
    }

#     set idx 1
#     for { set i 0 } { $i < [llength $coordList] } { incr i 2 } {
#         lassign [lrange $coordList $i $i+1] x y
#         CreateCircle $x $y 2 1 green 0 1
#         WriteText $idx [expr {$x+2}] [expr {$y+2}] Arial 12 sw green
#         incr idx
#     }
    
    _ResetColorDashLineWidth
}

proc PDFWriter::PDF::CreateCircle { x y radius lwidth color { dash 0 } { isfilled 0 } } {
    variable pdfhandle
    variable pagehandle

    _SetColorDashLineWidth $color $dash $lwidth

    $pagehandle circle $x $y $radius
    if { !$isfilled } {
	$pagehandle stroke
    } else {
	$pagehandle fill
    }
    _ResetColorDashLineWidth
}

# arcs begin in east and are counterclockwise (in degrees)
# isfilled can be: 0 (arc), 1 (chord) or pieslice
proc PDFWriter::PDF::CreateArc { x y radius Sangle Eangle lwidth color { dash 0 } { isfilled 0 } } {
    variable pdfhandle
    variable pagehandle

    math::constants::constants degtorad

    _SetColorDashLineWidth $color $dash $lwidth
    
    set Sangle_ $Sangle
    set Eangle_ $Eangle

    # angles begin up and go clockwise
    foreach i [list Sangle Eangle] {
	set $i [expr {360-([set $i]-90)}]
    }
    if { $Sangle-$Eangle >= 360 } {
	set Sangle [expr {$Eangle+359.9}]
    }
    if { $isfilled eq "pieslice" } {
	$pagehandle moveTo $x $y
	$pagehandle lineTo [expr {$x+$radius*cos($degtorad*$Sangle_)}] \
	    [expr {$y+$radius*sin($degtorad*$Sangle_)}]
    }
    
    $pagehandle arc $x $y $radius $Eangle $Sangle
    
    if { $isfilled eq "pieslice" } {
	$pagehandle lineTo $x $y
	$pagehandle closePath
    }
    
    if { $isfilled == 0 } {
	$pagehandle stroke
    } else {
	$pagehandle fill
    }
    
    _ResetColorDashLineWidth
}

proc PDFWriter::PDF::CreateEllipse { x y radiusx radiusy Sangle Eangle lwidth color { dash 0 } { isfilled 0 } } {
    
    ScaleCoordSystem $x $y 1.0 [expr {$radiusy/double($radiusx)}]
    CreateArc $x $y $radiusx $Sangle $Eangle $lwidth $color $dash $isfilled
    RestoreCoordSystem
}

proc PDFWriter::PDF::CreatePolygon { coordList color { isfilled 0 } { lwidth "" } { dash 0 } } {
    variable pdfhandle
    variable pagehandle
    
    _SetColorDashLineWidth $color $dash $lwidth

    $pagehandle moveTo {*}[lrange $coordList 0 1]

    foreach "x y" [lrange $coordList 2 end] {
	$pagehandle lineTo $x $y
    }
    $pagehandle closePath

    if { !$isfilled } {
	$pagehandle stroke
    } else {
	$pagehandle fill
    }
    _ResetColorDashLineWidth
}

proc PDFWriter::PDF::CreateFilledBox { x y width height anchor lwidth colorL colorF { dash 0 } } {
    variable pdfhandle
    variable pagehandle

    if { [llength $colorF] == 2 } {
	lassign $colorF colorF rounded
    } else {
	set rounded ""
    }
    if { $anchor eq "center" } { set anchor "" }
    if { [string first n $anchor] != -1 } {
	set y [expr $y-$height]
    } elseif { [string first s $anchor] != -1 } {
	# nothing
    } else {
	set y [expr $y-$height/2.0]
    }
    if { [string first w $anchor] != -1 } {
	# nothing
    } elseif { [string first e $anchor] != -1 } {
	set x [expr $x-$width]
    } else {
	set x [expr $x-$width/2.0]
    }
    foreach i "F L" {
	if { [set color$i] == "" } { continue }
	if { $i eq "L" && ($lwidth eq "" || $lwidth == 0) } { continue }
	_SetColorDashLineWidth [set color$i] $dash $lwidth
	$pagehandle moveTo $x $y
	if { [string first e $rounded] != -1 } {
	    set r [expr {.5*$height}]
	    $pagehandle lineTo [expr $x+$width-2*$r] $y
	    $pagehandle curveTo [expr $x+$width] $y [expr $x+$width] [expr $y+$height] \
		[expr $x+$width-2*$r] [expr $y+$height]
	} else {
	    $pagehandle lineTo [expr $x+$width] $y
	    $pagehandle lineTo [expr $x+$width] [expr $y+$height]
	}
	$pagehandle lineTo $x [expr $y+$height]
	switch $i {
	    L { $pagehandle closePathStroke }
	    F { $pagehandle fill }
	}
    }
    _ResetColorDashLineWidth
}

proc PDFWriter::PDF::DrawGrid { { unit cm } } {
    variable xmax
    variable ymax

    switch $unit {
	cm { set delta [cm2pt 1] }
	pt {  set delta 50 }
    }
    set width 0.1
    
    for { set x 0 } { $x < $xmax } { set x [expr {$x+$delta}] } {
	PDFWriter::PDF::CreateLine $x 0 $x $ymax $width black
	PDFWriter::PDF::CreateLine [expr {$x+0.5*$delta}] 0 [expr {$x+0.5*$delta}] $ymax $width black 1
	switch $unit {
	    cm { set text [expr int([pt2cm $x])] }
	    pt {  set text $x }
	}
	WriteText $text $x [expr {$ymax-[cm2pt 1]}] Helvetica 10 nw
	WriteText $text $x [cm2pt 1] Helvetica 10 nw
    }
    for { set y 0 } { $y < $ymax } { set y [expr {$y+$delta}] } {
	PDFWriter::PDF::CreateLine 0 $y $xmax $y $width black
	PDFWriter::PDF::CreateLine 0 [expr {$y+0.5*$delta}] $xmax [expr {$y+0.5*$delta}] $width black 1
	switch $unit {
	    cm { set text [expr int([pt2cm $y])] }
	    pt {  set text $y }
	}
	WriteText $text [expr  {$xmax-[cm2pt 1]}] $y Helvetica 10 nw
	WriteText $text [cm2pt 1] $y Helvetica 10 nw
    }
}

proc PDFWriter::PDF::LineSpace { font size } {
    variable pdfhandle
    variable pagehandle
    variable fonts

    SetFontAndSize $font $size
    set leading [$pagehandle getTextLeading]
    #set leading 0
    if { $leading == 0 } { set leading $size }
    return [expr $leading*1.2]
}

proc PDFWriter::PDF::CapHeight { font size } {
    variable pdfhandle
    variable fonts

    SetFontAndSize $font $size
    set s [expr {[$fonts($font) getCapHeight]/1000.0}]
    if { $s == 0 } {
	set s [expr {[$fonts(Helvetica) getCapHeight]/1000.0}]
    }
    return $s
}

proc PDFWriter::PDF::XHeight { font size } {
    variable pdfhandle
    variable fonts

    SetFontAndSize $font $size
    set s [expr {[$fonts($font) getXHeight]/1000.0}]
    if { $s == 0 } {
	set s [expr {[$fonts(Helvetica) getXHeight]/1000.0}]
    }
    return $s
}

proc PDFWriter::PDF::Ascender { font size } {
    variable pdfhandle
    variable fonts
    
    SetFontAndSize $font $size
    set s [expr {[$fonts($font) getAscent]/1000.0}]
    if { $s == 0 } {
	set s [expr {[$fonts(Helvetica) getAscent]/1000.0}]
    }
    return $s
}

proc PDFWriter::PDF::Descender { font size } {
    variable pdfhandle
    variable fonts

    SetFontAndSize $font $size
    set s [expr {[$fonts($font) getDescent]/1000.0}]
    if { $s == 0 } {
	set s [expr {[$fonts(Helvetica) getDescent]/1000.0}]
    }
    return $s
}

proc PDFWriter::PDF::TextWidth { text font size } {
    variable pdfhandle
    variable pagehandle
    variable fonts

    SetFontAndSize $font $size
    return [$pagehandle textWidth [_encode_text $text $font]]
}

proc PDFWriter::PDF::_encode_text { text font } {
    variable tcl_encoding

    if { $font eq "Symbol" } {
	return [encoding convertto cp1252 $text]
    } else {
	return [encoding convertto $tcl_encoding $text]
    }
}

proc PDFWriter::PDF::WriteText { text x y font size anchor { color "" } } {
    variable pdfhandle
    variable pagehandle
    variable fonts

    
    if 0 {
	_SetColorDashLineWidth blue "" ""
	$pagehandle circle $x $y .5
	$pagehandle fill
	SetFontAndSize $font 6
	$pagehandle beginText
	$pagehandle textOut [expr {$x+2}] $y [_encode_text $anchor $font]
	$pagehandle endText
	_ResetColorDashLineWidth
    }
    _SetColorDashLineWidth $color "" ""
    SetFontAndSize $font $size

    set width 0
    set NL 0
    set height1 [$pagehandle getTextLeading]
    if { $height1 == 0 } { set height1 $size }
    set height2 [expr {[Ascender $font $size]*$size}]

    foreach i [::textutil::split::splitx $text {\s*(?:\n|\\n)\s*}] {
	if { [$pagehandle textWidth [_encode_text $i $font]] > $width } {
	    set width [$pagehandle textWidth [_encode_text $i $font]]
	}
	incr NL
    }
    set height [expr ($NL-1)*1.2*$height1+$height2]
    
    if { $anchor eq "center" } { set anchor "" }
    if { [string first n $anchor] != -1 } {
	set y [expr $y-$height]
    } elseif { [string first s $anchor] != -1 } {
	# nothing
    } else {
	set y [expr $y-$height/2.0]
    }
    if { [string first w $anchor] != -1 } {
	# nothing
    } elseif { [string first e $anchor] != -1 } {
	set x [expr $x-$width]
    } else {
	set x [expr $x-$width/2.0]
    }

    #PDFWriter::CreateLine $x $y [expr {$x+10}] $y 0.1 green

    $pagehandle beginText

    foreach i [::textutil::split::splitx $text {\s*(?:\n|\\n)\s*}] {
	$pagehandle textOut $x $y [_encode_text $i $font]
	set y [expr $y-1.2*$height1]
    }
    $pagehandle endText
    _ResetColorDashLineWidth


#     _SetColorDashLineWidth red "" ""
#     PDF_circle $pdfhandle $x [expr $y+1.2*$height1] .5
#     PDF_fill $pdfhandle
#     _ResetColorDashLineWidth


    return [list $width [expr $NL*1.2*$height1]]
}

proc PDFWriter::PDF::ScaleCoordSystem { xc yc scalefactx { scalefacty "" } } {
    variable pdfhandle
    variable pagehandle

    if { $scalefacty eq "" } { set scalefacty $scalefactx }

    $pagehandle gSave
    $pagehandle concat $scalefactx 0 0 $scalefacty $xc $yc
    $pagehandle concat 1 0 0 1 [expr {-1*$xc}] [expr {-1*$yc}]
}

proc PDFWriter::PDF::RotateCoordSystem { x y angle } {
    variable pdfhandle
    variable pagehandle
    
    math::constants::constants degtorad
    set angle [expr {$angle*$degtorad}]

    $pagehandle gSave
    $pagehandle concat [expr {cos($angle)}] [expr {sin($angle)}] \
	[expr {-sin($angle)}] [expr {cos($angle)}] \
	$x $y
    $pagehandle concat 1 0 0 1 [expr {-1*$x}] [expr {-1*$y}]
}

proc PDFWriter::PDF::RestoreCoordSystem {} {
    variable pdfhandle
    variable pagehandle

    $pagehandle gRestore
}

# angle in degrees counterclockwise from the positive x axis
proc PDFWriter::PDF::WriteTextRotated { text x y font size anchor color angle } {
    variable pdfhandle
    variable pagehandle
    
    math::constants::constants degtorad
    set angle [expr {$angle*$degtorad}]

    $pagehandle gSave
    $pagehandle concat [expr {cos($angle)}] [expr {sin($angle)}] \
	[expr {-sin($angle)}] [expr {cos($angle)}] \
	$x $y
    WriteText $text 0 0 $font $size $anchor $color
    $pagehandle gRestore
}

# mode can be left, right, center, justify or fulljustify
proc PDFWriter::PDF::WriteTextBox { text x y width mode font size } {
    variable pdfhandle
    variable pagehandle
    variable fonts

    SetFontAndSize $font $size

    if { $width == 0 } {
	switch $mode {
	    left { set anchor w }
	    right { set anchor e }
	    center { set anchor center }
	}
	WriteText $text $x $y $font $size $anchor
	return
    }
    set x [expr $x+[cm2pt 0.25]]
    set width [expr $width-2*[cm2pt 0.25]]

    set heightbase [$pagehandle getTextLeading]
    #set heightbase 0
    if { $heightbase == 0 } { set heightbase $size }
    set height $heightbase
    set textL $text
    
    while 1 {
	set r [$pagehandle measureText [_encode_text $textL $font] $width 0]
	if { $r == [string length $textL] } { break }
	set height [expr $height+$heightbase]
	set textL [string range $textL $r end]
    }
    set heightT [expr {$height+0.3*$heightbase}]
    $pagehandle beginText
    $pagehandle textRect $x [expr {$y+$heightbase/2}] [expr {$x+$width}] [expr $y-$heightT] \
	[_encode_text $text $font] $mode
    $pagehandle endText
    return [expr $y-$height+$heightbase/2]
}

proc PDFWriter::PDF::LocalLink { llx lly urx ury page_num } {
    variable local_link_stack
    variable pagehandle
    
    lappend local_link_stack [list $pagehandle $llx $lly $urx $ury $page_num]
}

proc PDFWriter::PDF::LocalLinkDo { pagehandleL llx lly urx ury page_num } {
    variable pdfhandle
    variable pagehandle

    set dest_pagehandle [$pdfhandle getPageByIndex [expr {$page_num-1}]]
    $pagehandleL createLinkAnnot [list $llx $lly $urx $ury] [$dest_pagehandle createDestination]
}

proc PDFWriter::PDF::End { { outputPDF 1 } } {
    variable pdffile
    variable pagehandle
    variable pdfhandle
    variable cacheimages
    variable local_link_stack
    
    array unset cacheimages

    if { $pdfhandle eq "" } {
	error "error in PDFWriter::PDF::End. There is no handle"
    }
    
    foreach i $local_link_stack {
	lassign $i pagehandleL llx lly urx ury page_num
	LocalLinkDo $pagehandleL $llx $lly $urx $ury $page_num
    }
    set local_link_stack ""
    
    $pdfhandle saveToFile $pdffile
    hpdf::free $pdfhandle
    set pdfhandle ""
    set pagehandle ""

    if { !$outputPDF } {
	return $pdffile
    }

    if { $::tcl_platform(platform) == "windows" } {
	set err [catch {
	    package require registry
	    set root HKEY_CLASSES_ROOT
	    set appKey [registry get $root\\.pdf ""]
	    set appCmd [registry get $root\\$appKey\\shell\\open\\command ""]
	    set appCmd [string map [list %1 [file native $pdffile]] $appCmd]
	    #regsub {%1} $appCmd [file native $pdffile] appCmd
	    regsub -all {\\} $appCmd {\\\\} appCmd
	    eval exec $appCmd &
	} errstring]
	if { $err } {
	    set file [tk_getSaveFile -filetypes \
		          {{{PDF files} ".pdf"} {{All files} "*"}}  \
		          -initialdir [pwd] \
		          -title "Save PDF file" -defaultextension .pdf]
	    if { $file == "" } { return }
	    if { [string equal [file normalize $pdffile] [file normalize $file]] } {
		return $pdffile
	    }
	    file copy -force $pdffile $file
	}
	#eval exec [auto_execok start] [list "" [file native $pdffile]] &
    } else {
	set programs [list xdg-open gnome-open acroread evince]
	if { $::tcl_platform(os) eq "Darwin" } {
	    set programs [linsert $programs 0 open]
	}
	set found 0
	foreach p $programs {
	    if { [auto_execok $p] ne "" } {
		exec {*}[auto_execok $p] $pdffile &
		set found 1
		break
	    }
	}
	if { !$found } {
	    WarnWin [_ "File cannot be visualized. Neither of programs '%s' is installed in the system" $programs]
	}
    }
    return $pdffile
}

set PDFWriter::__tcl3dColorList {
    "alice blue"               240   248   255
    "AliceBlue"                240   248   255
    "antique white"            250   235   215
    "AntiqueWhite"             250   235   215
    "AntiqueWhite1"            255   239   219
    "AntiqueWhite2"            238   223   204
    "AntiqueWhite3"            205   192   176
    "AntiqueWhite4"            139   131   120
    "aquamarine"               127   255   212
    "aquamarine1"              127   255   212
    "aquamarine2"              118   238   198
    "aquamarine3"              102   205   170
    "aquamarine4"               69   139   116
    "azure"                    240   255   255
    "azure1"                   240   255   255
    "azure2"                   224   238   238
    "azure3"                   193   205   205
    "azure4"                   131   139   139
    "beige"                    245   245   220
    "bisque"                   255   228   196
    "bisque1"                  255   228   196
    "bisque2"                  238   213   183
    "bisque3"                  205   183   158
    "bisque4"                  139   125   107
    "blanched almond"          255   235   205
    "BlanchedAlmond"           255   235   205
    "blue violet"              138    43   226
    "blue1"                      0     0   255
    "blue2"                      0     0   238
    "blue3"                      0     0   205
    "blue4"                      0     0   139
    "BlueViolet"               138    43   226
    "brown"                    165    42    42
    "brown1"                   255    64    64
    "brown2"                   238    59    59
    "brown3"                   205    51    51
    "brown4"                   139    35    35
    "burlywood"                222   184   135
    "burlywood1"               255   211   155
    "burlywood2"               238   197   145
    "burlywood3"               205   170   125
    "burlywood4"               139   115    85
    "cadet blue"                95   158   160
    "CadetBlue"                 95   158   160
    "CadetBlue1"               152   245   255
    "CadetBlue2"               142   229   238
    "CadetBlue3"               122   197   205
    "CadetBlue4"                83   134   139
    "chartreuse"               127   255     0
    "chartreuse1"              127   255     0
    "chartreuse2"              118   238     0
    "chartreuse3"              102   205     0
    "chartreuse4"               69   139     0
    "chocolate"                210   105    30
    "chocolate1"               255   127    36
    "chocolate2"               238   118    33
    "chocolate3"               205   102    29
    "chocolate4"               139    69    19
    "coral"                    255   127    80
    "coral1"                   255   114    86
    "coral2"                   238   106    80
    "coral3"                   205    91    69
    "coral4"                   139    62    47
    "cornflower blue"          100   149   237
    "CornflowerBlue"           100   149   237
    "cornsilk1"                255   248   220
    "cornsilk2"                238   232   205
    "cornsilk3"                205   200   177
    "cornsilk4"                139   136   120
    "cornsilk"                 255   248   220
    "cyan1"                      0   255   255
    "cyan2"                      0   238   238
    "cyan3"                      0   205   205
    "cyan4"                      0   139   139
    "dark blue"                  0     0   139
    "dark cyan"                  0   139   139
    "dark goldenrod"           184   134    11
    "dark gray"                169   169   169
    "dark green"                 0   100     0
    "dark grey"                169   169   169
    "dark khaki"               189   183   107
    "dark magenta"             139     0   139
    "dark olive green"          85   107    47
    "dark orange"              255   140     0
    "dark orchid"              153    50   204
    "dark red"                 139     0     0
    "dark salmon"              233   150   122
    "dark sea green"           143   188   143
    "dark slate blue"           72    61   139
    "dark slate gray"           47    79    79
    "dark slate grey"           47    79    79
    "dark turquoise"             0   206   209
    "dark violet"              148     0   211
    "DarkBlue"                   0     0   139
    "DarkCyan"                   0   139   139
    "DarkGoldenrod"            184   134    11
    "DarkGoldenrod1"           255   185    15
    "DarkGoldenrod2"           238   173    14
    "DarkGoldenrod3"           205   149    12
    "DarkGoldenrod4"           139   101     8
    "DarkGray"                 169   169   169
    "DarkGreen"                  0   100     0
    "DarkGrey"                 169   169   169
    "DarkKhaki"                189   183   107
    "DarkMagenta"              139     0   139
    "DarkOliveGreen"            85   107    47
    "DarkOliveGreen1"          202   255   112
    "DarkOliveGreen2"          188   238   104
    "DarkOliveGreen3"          162   205    90
    "DarkOliveGreen4"          110   139    61
    "DarkOrange"               255   140     0
    "DarkOrange1"              255   127     0
    "DarkOrange2"              238   118     0
    "DarkOrange3"              205   102     0
    "DarkOrange4"              139    69     0
    "DarkOrchid"               153    50   204
    "DarkOrchid1"              191    62   255
    "DarkOrchid2"              178    58   238
    "DarkOrchid3"              154    50   205
    "DarkOrchid4"              104    34   139
    "DarkRed"                  139     0     0
    "DarkSalmon"               233   150   122
    "DarkSeaGreen"             143   188   143
    "DarkSeaGreen1"            193   255   193
    "DarkSeaGreen2"            180   238   180
    "DarkSeaGreen3"            155   205   155
    "DarkSeaGreen4"            105   139   105
    "DarkSlateBlue"             72    61   139
    "DarkSlateGray1"           151   255   255
    "DarkSlateGray2"           141   238   238
    "DarkSlateGray3"           121   205   205
    "DarkSlateGray4"            82   139   139
    "DarkSlateGray"             47    79    79
    "DarkSlateGrey"             47    79    79
    "DarkTurquoise"              0   206   209
    "DarkViolet"               148     0   211
    "deep pink"                255    20   147
    "deep sky blue"              0   191   255
    "DeepPink1"                255    20   147
    "DeepPink2"                238    18   137
    "DeepPink3"                205    16   118
    "DeepPink4"                139    10    80
    "DeepPink"                 255    20   147
    "DeepSkyBlue1"               0   191   255
    "DeepSkyBlue2"               0   178   238
    "DeepSkyBlue3"               0   154   205
    "DeepSkyBlue4"               0   104   139
    "DeepSkyBlue"                0   191   255
    "dim gray"                 105   105   105
    "dim grey"                 105   105   105
    "DimGray"                  105   105   105
    "DimGrey"                  105   105   105
    "dodger blue"               30   144   255
    "DodgerBlue1"               30   144   255
    "DodgerBlue2"               28   134   238
    "DodgerBlue3"               24   116   205
    "DodgerBlue4"               16    78   139
    "DodgerBlue"                30   144   255
    "firebrick1"               255    48    48
    "firebrick2"               238    44    44
    "firebrick3"               205    38    38
    "firebrick4"               139    26    26
    "firebrick"                178    34    34
    "floral white"             255   250   240
    "FloralWhite"              255   250   240
    "forest green"              34   139    34
    "ForestGreen"               34   139    34
    "gainsboro"                220   220   220
    "ghost white"              248   248   255
    "GhostWhite"               248   248   255
    "gold1"                    255   215     0
    "gold2"                    238   201     0
    "gold3"                    205   173     0
    "gold4"                    139   117     0
    "gold"                     255   215     0
    "goldenrod1"               255   193    37
    "goldenrod2"               238   180    34
    "goldenrod3"               205   155    29
    "goldenrod4"               139   105    20
    "goldenrod"                218   165    32
    "gray"                     190   190   190
    "gray0"                      0     0     0
    "gray1"                      3     3     3
    "gray2"                      5     5     5
    "gray3"                      8     8     8
    "gray4"                     10    10    10
    "gray5"                     13    13    13
    "gray6"                     15    15    15
    "gray7"                     18    18    18
    "gray8"                     20    20    20
    "gray9"                     23    23    23
    "gray10"                    26    26    26
    "gray11"                    28    28    28
    "gray12"                    31    31    31
    "gray13"                    33    33    33
    "gray14"                    36    36    36
    "gray15"                    38    38    38
    "gray16"                    41    41    41
    "gray17"                    43    43    43
    "gray18"                    46    46    46
    "gray19"                    48    48    48
    "gray20"                    51    51    51
    "gray21"                    54    54    54
    "gray22"                    56    56    56
    "gray23"                    59    59    59
    "gray24"                    61    61    61
    "gray25"                    64    64    64
    "gray26"                    66    66    66
    "gray27"                    69    69    69
    "gray28"                    71    71    71
    "gray29"                    74    74    74
    "gray30"                    77    77    77
    "gray31"                    79    79    79
    "gray32"                    82    82    82
    "gray33"                    84    84    84
    "gray34"                    87    87    87
    "gray35"                    89    89    89
    "gray36"                    92    92    92
    "gray37"                    94    94    94
    "gray38"                    97    97    97
    "gray39"                    99    99    99
    "gray40"                   102   102   102
    "gray41"                   105   105   105
    "gray42"                   107   107   107
    "gray43"                   110   110   110
    "gray44"                   112   112   112
    "gray45"                   115   115   115
    "gray46"                   117   117   117
    "gray47"                   120   120   120
    "gray48"                   122   122   122
    "gray49"                   125   125   125
    "gray50"                   127   127   127
    "gray51"                   130   130   130
    "gray52"                   133   133   133
    "gray53"                   135   135   135
    "gray54"                   138   138   138
    "gray55"                   140   140   140
    "gray56"                   143   143   143
    "gray57"                   145   145   145
    "gray58"                   148   148   148
    "gray59"                   150   150   150
    "gray60"                   153   153   153
    "gray61"                   156   156   156
    "gray62"                   158   158   158
    "gray63"                   161   161   161
    "gray64"                   163   163   163
    "gray65"                   166   166   166
    "gray66"                   168   168   168
    "gray67"                   171   171   171
    "gray68"                   173   173   173
    "gray69"                   176   176   176
    "gray70"                   179   179   179
    "gray71"                   181   181   181
    "gray72"                   184   184   184
    "gray73"                   186   186   186
    "gray74"                   189   189   189
    "gray75"                   191   191   191
    "gray76"                   194   194   194
    "gray77"                   196   196   196
    "gray78"                   199   199   199
    "gray79"                   201   201   201
    "gray80"                   204   204   204
    "gray81"                   207   207   207
    "gray82"                   209   209   209
    "gray83"                   212   212   212
    "gray84"                   214   214   214
    "gray85"                   217   217   217
    "gray86"                   219   219   219
    "gray87"                   222   222   222
    "gray88"                   224   224   224
    "gray89"                   227   227   227
    "gray90"                   229   229   229
    "gray91"                   232   232   232
    "gray92"                   235   235   235
    "gray93"                   237   237   237
    "gray94"                   240   240   240
    "gray95"                   242   242   242
    "gray96"                   245   245   245
    "gray97"                   247   247   247
    "gray98"                   250   250   250
    "gray99"                   252   252   252
    "gray100"                  255   255   255
    "green yellow"             173   255    47
    "green1"                     0   255     0
    "green2"                     0   238     0
    "green3"                     0   205     0
    "green4"                     0   139     0
    "GreenYellow"              173   255    47
    "grey0"                      0     0     0
    "grey1"                      3     3     3
    "grey2"                      5     5     5
    "grey3"                      8     8     8
    "grey4"                     10    10    10
    "grey5"                     13    13    13
    "grey6"                     15    15    15
    "grey7"                     18    18    18
    "grey8"                     20    20    20
    "grey9"                     23    23    23
    "grey10"                    26    26    26
    "grey11"                    28    28    28
    "grey12"                    31    31    31
    "grey13"                    33    33    33
    "grey14"                    36    36    36
    "grey15"                    38    38    38
    "grey16"                    41    41    41
    "grey17"                    43    43    43
    "grey18"                    46    46    46
    "grey19"                    48    48    48
    "grey20"                    51    51    51
    "grey21"                    54    54    54
    "grey22"                    56    56    56
    "grey23"                    59    59    59
    "grey24"                    61    61    61
    "grey25"                    64    64    64
    "grey26"                    66    66    66
    "grey27"                    69    69    69
    "grey28"                    71    71    71
    "grey29"                    74    74    74
    "grey30"                    77    77    77
    "grey31"                    79    79    79
    "grey32"                    82    82    82
    "grey33"                    84    84    84
    "grey34"                    87    87    87
    "grey35"                    89    89    89
    "grey36"                    92    92    92
    "grey37"                    94    94    94
    "grey38"                    97    97    97
    "grey39"                    99    99    99
    "grey40"                   102   102   102
    "grey41"                   105   105   105
    "grey42"                   107   107   107
    "grey43"                   110   110   110
    "grey44"                   112   112   112
    "grey45"                   115   115   115
    "grey46"                   117   117   117
    "grey47"                   120   120   120
    "grey48"                   122   122   122
    "grey49"                   125   125   125
    "grey50"                   127   127   127
    "grey51"                   130   130   130
    "grey52"                   133   133   133
    "grey53"                   135   135   135
    "grey54"                   138   138   138
    "grey55"                   140   140   140
    "grey56"                   143   143   143
    "grey57"                   145   145   145
    "grey58"                   148   148   148
    "grey59"                   150   150   150
    "grey60"                   153   153   153
    "grey61"                   156   156   156
    "grey62"                   158   158   158
    "grey63"                   161   161   161
    "grey64"                   163   163   163
    "grey65"                   166   166   166
    "grey66"                   168   168   168
    "grey67"                   171   171   171
    "grey68"                   173   173   173
    "grey69"                   176   176   176
    "grey70"                   179   179   179
    "grey71"                   181   181   181
    "grey72"                   184   184   184
    "grey73"                   186   186   186
    "grey74"                   189   189   189
    "grey75"                   191   191   191
    "grey76"                   194   194   194
    "grey77"                   196   196   196
    "grey78"                   199   199   199
    "grey79"                   201   201   201
    "grey80"                   204   204   204
    "grey81"                   207   207   207
    "grey82"                   209   209   209
    "grey83"                   212   212   212
    "grey84"                   214   214   214
    "grey85"                   217   217   217
    "grey86"                   219   219   219
    "grey87"                   222   222   222
    "grey88"                   224   224   224
    "grey89"                   227   227   227
    "grey90"                   229   229   229
    "grey91"                   232   232   232
    "grey92"                   235   235   235
    "grey93"                   237   237   237
    "grey94"                   240   240   240
    "grey95"                   242   242   242
    "grey96"                   245   245   245
    "grey97"                   247   247   247
    "grey98"                   250   250   250
    "grey99"                   252   252   252
    "grey100"                  255   255   255
    "grey"                     190   190   190
    "honeydew1"                240   255   240
    "honeydew2"                224   238   224
    "honeydew3"                193   205   193
    "honeydew4"                131   139   131
    "honeydew"                 240   255   240
    "hot pink"                 255   105   180
    "HotPink1"                 255   110   180
    "HotPink2"                 238   106   167
    "HotPink3"                 205    96   144
    "HotPink4"                 139    58    98
    "HotPink"                  255   105   180
    "indian red"               205    92    92
    "IndianRed1"               255   106   106
    "IndianRed2"               238    99    99
    "IndianRed3"               205    85    85
    "IndianRed4"               139    58    58
    "IndianRed"                205    92    92
    "ivory1"                   255   255   240
    "ivory2"                   238   238   224
    "ivory3"                   205   205   193
    "ivory4"                   139   139   131
    "ivory"                    255   255   240
    "khaki1"                   255   246   143
    "khaki2"                   238   230   133
    "khaki3"                   205   198   115
    "khaki4"                   139   134    78
    "khaki"                    240   230   140
    "lavender"                 230   230   250
    "lavender blush"           255   240   245
    "LavenderBlush1"           255   240   245
    "LavenderBlush2"           238   224   229
    "LavenderBlush3"           205   193   197
    "LavenderBlush4"           139   131   134
    "LavenderBlush"            255   240   245
    "lawn green"               124   252     0
    "LawnGreen"                124   252     0
    "lemon chiffon"            255   250   205
    "LemonChiffon1"            255   250   205
    "LemonChiffon2"            238   233   191
    "LemonChiffon3"            205   201   165
    "LemonChiffon4"            139   137   112
    "LemonChiffon"             255   250   205
    "light blue"               173   216   230
    "light coral"              240   128   128
    "light cyan"               224   255   255
    "light goldenrod"          238   221   130
    "light goldenrod yellow"   250   250   210
    "light gray"               211   211   211
    "light green"              144   238   144
    "light grey"               211   211   211
    "light pink"               255   182   193
    "light salmon"             255   160   122
    "light sea green"           32   178   170
    "light sky blue"           135   206   250
    "light slate blue"         132   112   255
    "light slate gray"         119   136   153
    "light slate grey"         119   136   153
    "light steel blue"         176   196   222
    "light yellow"             255   255   224
    "LightBlue1"               191   239   255
    "LightBlue2"               178   223   238
    "LightBlue3"               154   192   205
    "LightBlue4"               104   131   139
    "LightBlue"                173   216   230
    "LightCoral"               240   128   128
    "LightCyan1"               224   255   255
    "LightCyan2"               209   238   238
    "LightCyan3"               180   205   205
    "LightCyan4"               122   139   139
    "LightCyan"                224   255   255
    "LightGoldenrod1"          255   236   139
    "LightGoldenrod2"          238   220   130
    "LightGoldenrod3"          205   190   112
    "LightGoldenrod4"          139   129    76
    "LightGoldenrod"           238   221   130
    "LightGoldenrodYellow"     250   250   210
    "LightGray"                211   211   211
    "LightGreen"               144   238   144
    "LightGrey"                211   211   211
    "LightPink1"               255   174   185
    "LightPink2"               238   162   173
    "LightPink3"               205   140   149
    "LightPink4"               139    95   101
    "LightPink"                255   182   193
    "LightSalmon1"             255   160   122
    "LightSalmon2"             238   149   114
    "LightSalmon3"             205   129    98
    "LightSalmon4"             139    87    66
    "LightSalmon"              255   160   122
    "LightSeaGreen"             32   178   170
    "LightSkyBlue1"            176   226   255
    "LightSkyBlue2"            164   211   238
    "LightSkyBlue3"            141   182   205
    "LightSkyBlue4"             96   123   139
    "LightSkyBlue"             135   206   250
    "LightSlateBlue"           132   112   255
    "LightSlateGray"           119   136   153
    "LightSlateGrey"           119   136   153
    "LightSteelBlue1"          202   225   255
    "LightSteelBlue2"          188   210   238
    "LightSteelBlue3"          162   181   205
    "LightSteelBlue4"          110   123   139
    "LightSteelBlue"           176   196   222
    "LightYellow1"             255   255   224
    "LightYellow2"             238   238   209
    "LightYellow3"             205   205   180
    "LightYellow4"             139   139   122
    "LightYellow"              255   255   224
    "lime green"                50   205    50
    "LimeGreen"                 50   205    50
    "linen"                    250   240   230
    "magenta1"                 255     0   255
    "magenta2"                 238     0   238
    "magenta3"                 205     0   205
    "magenta4"                 139     0   139
    "magenta"                  255     0   255
    "maroon1"                  255    52   179
    "maroon2"                  238    48   167
    "maroon3"                  205    41   144
    "maroon4"                  139    28    98
    "maroon"                   176    48    96
    "medium aquamarine"        102   205   170
    "medium blue"                0     0   205
    "medium orchid"            186    85   211
    "medium purple"            147   112   219
    "medium sea green"          60   179   113
    "medium slate blue"        123   104   238
    "medium spring green"        0   250   154
    "medium turquoise"          72   209   204
    "medium violet red"        199    21   133
    "MediumAquamarine"         102   205   170
    "MediumBlue"                 0     0   205
    "MediumOrchid1"            224   102   255
    "MediumOrchid2"            209    95   238
    "MediumOrchid3"            180    82   205
    "MediumOrchid4"            122    55   139
    "MediumOrchid"             186    85   211
    "MediumPurple1"            171   130   255
    "MediumPurple2"            159   121   238
    "MediumPurple3"            137   104   205
    "MediumPurple4"             93    71   139
    "MediumPurple"             147   112   219
    "MediumSeaGreen"            60   179   113
    "MediumSlateBlue"          123   104   238
    "MediumSpringGreen"          0   250   154
    "MediumTurquoise"           72   209   204
    "MediumVioletRed"          199    21   133
    "midnight blue"             25    25   112
    "MidnightBlue"              25    25   112
    "mint cream"               245   255   250
    "MintCream"                245   255   250
    "misty rose"               255   228   225
    "MistyRose1"               255   228   225
    "MistyRose2"               238   213   210
    "MistyRose3"               205   183   181
    "MistyRose4"               139   125   123
    "MistyRose"                255   228   225
    "moccasin"                 255   228   181
    "navajo white"             255   222   173
    "NavajoWhite1"             255   222   173
    "NavajoWhite2"             238   207   161
    "NavajoWhite3"             205   179   139
    "NavajoWhite4"             139   121    94
    "NavajoWhite"              255   222   173
    "navy"                       0     0   128
    "navy blue"                  0     0   128
    "NavyBlue"                   0     0   128
    "old lace"                 253   245   230
    "OldLace"                  253   245   230
    "olive drab"               107   142    35
    "OliveDrab1"               192   255    62
    "OliveDrab2"               179   238    58
    "OliveDrab3"               154   205    50
    "OliveDrab4"               105   139    34
    "OliveDrab"                107   142    35
    "orange red"               255    69     0
    "orange1"                  255   165     0
    "orange2"                  238   154     0
    "orange3"                  205   133     0
    "orange4"                  139    90     0
    "orange"                   255   165     0
    "OrangeRed1"               255    69     0
    "OrangeRed2"               238    64     0
    "OrangeRed3"               205    55     0
    "OrangeRed4"               139    37     0
    "OrangeRed"                255    69     0
    "orchid1"                  255   131   250
    "orchid2"                  238   122   233
    "orchid3"                  205   105   201
    "orchid4"                  139    71   137
    "orchid"                   218   112   214
    "pale goldenrod"           238   232   170
    "pale green"               152   251   152
    "pale turquoise"           175   238   238
    "pale violet red"          219   112   147
    "PaleGoldenrod"            238   232   170
    "PaleGreen1"               154   255   154
    "PaleGreen2"               144   238   144
    "PaleGreen3"               124   205   124
    "PaleGreen4"                84   139    84
    "PaleGreen"                152   251   152
    "PaleTurquoise1"           187   255   255
    "PaleTurquoise2"           174   238   238
    "PaleTurquoise3"           150   205   205
    "PaleTurquoise4"           102   139   139
    "PaleTurquoise"            175   238   238
    "PaleVioletRed1"           255   130   171
    "PaleVioletRed2"           238   121   159
    "PaleVioletRed3"           205   104   127
    "PaleVioletRed4"           139    71    93
    "PaleVioletRed"            219   112   147
    "papaya whip"              255   239   213
    "PapayaWhip"               255   239   213
    "peach puff"               255   218   185
    "PeachPuff1"               255   218   185
    "PeachPuff2"               238   203   173
    "PeachPuff3"               205   175   149
    "PeachPuff4"               139   119   101
    "PeachPuff"                255   218   185
    "peru"                     205   133    63
    "pink1"                    255   181   197
    "pink2"                    238   169   184
    "pink3"                    205   145   158
    "pink4"                    139    99   108
    "pink"                     255   192   203
    "plum1"                    255   187   255
    "plum2"                    238   174   238
    "plum3"                    205   150   205
    "plum4"                    139   102   139
    "plum"                     221   160   221
    "powder blue"              176   224   230
    "PowderBlue"               176   224   230
    "purple1"                  155    48   255
    "purple2"                  145    44   238
    "purple3"                  125    38   205
    "purple4"                   85    26   139
    "purple"                   160    32   240
    "red1"                     255     0     0
    "red2"                     238     0     0
    "red3"                     205     0     0
    "red4"                     139     0     0
    "rosy brown"               188   143   143
    "RosyBrown1"               255   193   193
    "RosyBrown2"               238   180   180
    "RosyBrown3"               205   155   155
    "RosyBrown4"               139   105   105
    "RosyBrown"                188   143   143
    "royal blue"                65   105   225
    "RoyalBlue1"                72   118   255
    "RoyalBlue2"                67   110   238
    "RoyalBlue3"                58    95   205
    "RoyalBlue4"                39    64   139
    "RoyalBlue"                 65   105   225
    "saddle brown"             139    69    19
    "SaddleBrown"              139    69    19
    "salmon1"                  255   140   105
    "salmon2"                  238   130    98
    "salmon3"                  205   112    84
    "salmon4"                  139    76    57
    "salmon"                   250   128   114
    "sandy brown"              244   164    96
    "SandyBrown"               244   164    96
    "sea green"                 46   139    87
    "SeaGreen1"                 84   255   159
    "SeaGreen2"                 78   238   148
    "SeaGreen3"                 67   205   128
    "SeaGreen4"                 46   139    87
    "SeaGreen"                  46   139    87
    "seashell1"                255   245   238
    "seashell2"                238   229   222
    "seashell3"                205   197   191
    "seashell4"                139   134   130
    "seashell"                 255   245   238
    "sienna1"                  255   130    71
    "sienna2"                  238   121    66
    "sienna3"                  205   104    57
    "sienna4"                  139    71    38
    "sienna"                   160    82    45
    "sky blue"                 135   206   235
    "SkyBlue1"                 135   206   255
    "SkyBlue2"                 126   192   238
    "SkyBlue3"                 108   166   205
    "SkyBlue4"                  74   112   139
    "SkyBlue"                  135   206   235
    "slate blue"               106    90   205
    "slate gray"               112   128   144
    "slate grey"               112   128   144
    "SlateBlue1"               131   111   255
    "SlateBlue2"               122   103   238
    "SlateBlue3"               105    89   205
    "SlateBlue4"                71    60   139
    "SlateBlue"                106    90   205
    "SlateGray1"               198   226   255
    "SlateGray2"               185   211   238
    "SlateGray3"               159   182   205
    "SlateGray4"               108   123   139
    "SlateGray"                112   128   144
    "SlateGrey"                112   128   144
    "snow1"                    255   250   250
    "snow2"                    238   233   233
    "snow3"                    205   201   201
    "snow4"                    139   137   137
    "snow"                     255   250   250
    "spring green"               0   255   127
    "SpringGreen1"               0   255   127
    "SpringGreen2"               0   238   118
    "SpringGreen3"               0   205   102
    "SpringGreen4"               0   139    69
    "SpringGreen"                0   255   127
    "steel blue"                70   130   180
    "SteelBlue1"                99   184   255
    "SteelBlue2"                92   172   238
    "SteelBlue3"                79   148   205
    "SteelBlue4"                54   100   139
    "SteelBlue"                 70   130   180
    "tan1"                     255   165    79
    "tan2"                     238   154    73
    "tan3"                     205   133    63
    "tan4"                     139    90    43
    "tan"                      210   180   140
    "thistle1"                 255   225   255
    "thistle2"                 238   210   238
    "thistle3"                 205   181   205
    "thistle4"                 139   123   139
    "thistle"                  216   191   216
    "tomato1"                  255    99    71
    "tomato2"                  238    92    66
    "tomato3"                  205    79    57
    "tomato4"                  139    54    38
    "tomato"                   255    99    71 
    "turquoise1"                 0   245   255
    "turquoise2"                 0   229   238
    "turquoise3"                 0   197   205
    "turquoise4"                 0   134   139
    "turquoise"                 64   224   208
    "violet red"               208    32   144
    "VioletRed1"               255    62   150
    "VioletRed2"               238    58   140
    "VioletRed3"               205    50   120
    "VioletRed4"               139    34    82
    "VioletRed"                208    32   144
    "wheat1"                   255   231   186
    "wheat2"                   238   216   174
    "wheat3"                   205   186   150
    "wheat4"                   139   126   102
    "wheat"                    245   222   179
    "white smoke"              245   245   245
    "WhiteSmoke"               245   245   245
    "yellow green"             154   205    50
    "yellow1"                  255   255     0
    "yellow2"                  238   238     0
    "yellow3"                  205   205     0
    "yellow4"                  139   139     0
    "YellowGreen"              154   205    50
    "blue"                       0     0   255
    "cyan"                       0   255   255
    "green"                      0   255     0
    "red"                      255     0     0
    "black"                      0     0     0
    "yellow"                   255   255     0
    "violet"                   238   130   238
    "white"                    255   255   255
}

